import struct
import os

# constant
SEG_SIZE = 32782
HEADER_SIZE = 12
DATA_SIZE = SEG_SIZE - HEADER_SIZE

class Segment():

    def __init__(self, seq_no=0, ack_no=0, data=bytes(), checksum=0):
        self.DATA_PTR = 12

        self.bytes = [0 for _ in range(HEADER_SIZE)]
        self.set_seq_no(seq_no)
        self.set_ack_no(ack_no)
        self.set_data(data)
        self.set_checksum(checksum)

    def set_seq_no(self, val):
        self.bytes[0] = val >> 24 & 0xff
        self.bytes[1] = val >> 16 & 0xff
        self.bytes[2] = val >> 8 & 0xff
        self.bytes[3] = val & 0xff

    def get_seq_no(self):
        return  (self.bytes[0] << 24) + \
                (self.bytes[1] << 16) + \
                (self.bytes[2] << 8) + \
                self.bytes[3]

    def set_ack_no(self, val):
        self.bytes[4] = val >> 24 & 0xff
        self.bytes[5] = val >> 16 & 0xff
        self.bytes[6] = val >> 8 & 0xff
        self.bytes[7] = val & 0xff

    def get_ack_no(self):
        return  (self.bytes[4] << 24) + \
                (self.bytes[5] << 16) + \
                (self.bytes[6] << 8) + \
                self.bytes[7]

    def set_data(self, data):
        ptr = self.DATA_PTR
        bound = len(data)
        if bound > DATA_SIZE:
            bound = DATA_SIZE
        for i in range(bound):
            self.bytes.append(data[i])
            
    def get_data(self):
        data=[]
        ptr = self.DATA_PTR   # pointer to data in segment
        for b in self.bytes[HEADER_SIZE:]:
            data.append(self.bytes[ptr])
            ptr += 1
        return bytes(data)

    def setSYN(self):
        # raise SYN flag
        self.bytes[8] = self.bytes[8] | 0b01000000
    
    def getSYN(self):
        return self.bytes[8] & 0b01000000
    
    def setACK(self):
        # raise ACK flag
        self.bytes[8] = self.bytes[8] | 0b00001000
    
    def getACK(self):
        return self.bytes[8] & 0b00001000

    def setFIN(self):
        # raise FIN flag
        self.bytes[8] = self.bytes[8] | 0b10000000

    def getFIN(self):
        return self.bytes[8] & 0b10000000

    def isData(self):
        return not self.bytes[8]

    def get_checksum(self):
        return  (self.bytes[10] << 8) + self.bytes[11]
    
    def set_checksum(self, val):
        self.bytes[10] = val >> 8 & 0xff
        self.bytes[11] = val & 0xff

    def count_checksum(self):
        byte10 = self.bytes[10]
        byte11 = self.bytes[11]
        self.bytes[10] = 0 # dangerous
        self.bytes[11] = 0 # dangerous
        data = self.to_bytes()

        len_data = len(data) # cara akses bagian data saja, karena perlu di rstrip("\x00")
        if (len_data % 2 != 0): # kalau panjangnya ganjil, kudu ditambahin 0 di akhir, pakai struct
            len_data += 1
            data += struct.pack('!B', 0)

        checksum=0
        for i in range (0, len_data, 2):
            w = (data[i] << 8) + (data[i + 1])
            checksum += w
            checksum = (checksum >> 16) + (checksum & 0xFFFF)

        self.bytes[10] = byte10 # dangerous
        self.bytes[11] = byte11 # dangerous

        return checksum

    def test_checksum(self):
        return self.count_checksum() == self.get_checksum()

    def to_bytes(self):
        return bytes(self.bytes)

    def from_bytes(self, byte:bytes):
        self.bytes = list(byte)
        return self

    def from_file(self, filepath):
        filename = os.path.basename(filepath)
        with open(filename, 'rb') as f:
            self.set_data(f.read())

    def to_file(self, filename):
        with open(filename, 'wb') as f:
            data = self.get_data()
            f.write(data)

    def log(self):
        print(self.bytes[:50])

if __name__=="__main__":
    seg = Segment(200, 199)
    seg.from_file("halo.txt")
    seg.set_checksum(seg.count_checksum())
    seg.log()

    file = open("halo.txt", 'r')
    seg.to_file("out.txt")